# Copyright 2021 Robin Scheibler
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import math
from typing import Optional, Tuple, Union

import numpy as np
from .helpers import clamp


def optimal_symmetric_circulant_precond_column(
    col_toeplitz: np.ndarray,
) -> np.ndarray:
    """
    compute the first column of the circulant matrix closest to
    the Toeplitz matrix with provided first column in terms of the
    Froebenius norm
    """
    n = col_toeplitz.shape[-1]
    b = np.arange(1, n) / n
    w = col_toeplitz[..., 1:] * b[..., ::-1]
    col_circ = np.concatenate((col_toeplitz[..., :1], w + w[..., ::-1]), axis=-1)
    return col_circ


def optimal_nonsymmetric_circulant_precond_column(
    r: np.ndarray, n: Optional[int] = None, axis: Optional[int] = None
) -> np.ndarray:
    """
    compute the first column of the circulant matrix closest to
    the non-symmetric Toeplitz matrix with provided first column in terms of the
    Froebenius norm

    The input is assumed to be of shape (..., 2 * n_col) in order
    r = [s_0, s_1, ..., s_{n-1}, (s_{-n}) , s_{-n+1}, ..., s_{-1}]

    where the first row of the Toeplitz matrix is
    [s_0, ..., s_{n-1}]
    and the first column is
    [s_0, s_{-1}, ..., s_{-n+1}]
    """
    if axis is not None:
        r = np.moveaxis(r, axis, -1)

    if n is None:
        n = (r.shape[-1] + 1) // 2

    b = np.arange(1, n) / n

    s1 = r[..., n - 1 : 0 : -1]
    s2 = r[..., :-n:-1]

    w = b * s1 + (1.0 - b) * s2

    c = np.concatenate((r[..., :1], w), axis=-1)

    if axis is not None:
        c = np.moveaxis(c, -1, axis)

    return c


class CirculantPreconditionerOperator:
    """
    Optimal circulant pre-conditioner operator for a symmetric topelitz matrix
    """

    def __init__(self, toeplitz_col: np.ndarray):
        col_precond = optimal_symmetric_circulant_precond_column(toeplitz_col)
        C = np.fft.rfft(col_precond, axis=-1)
        # complex pointwise inverse
        self.C = C.conj() / clamp(C.real**2 + C.imag**2, min=1e-6)
        self.n = col_precond.shape[-1]

        self._shape = col_precond.shape[:-1] + (self.n, self.n)

    @property
    def shape(self):
        return self._shape

    @property
    def ndim(self):
        return len(self._shape)

    def __matmul__(self, lhs: np.ndarray) -> np.ndarray:

        if lhs.ndim == self.ndim:
            assert lhs.shape[-2] == self.shape[-1]
            return np.fft.irfft(
                self.C[..., None] * np.fft.rfft(lhs, n=self.n, axis=-2),
                n=self.n,
                axis=-2,
            )

        else:
            raise ValueError(
                "Dimension mismatch between operators with "
                f"shapes {self.shape} and {lhs.shape}"
            )


class SymmetricToeplitzOperator:
    def __init__(self, col: np.ndarray):
        """
        col: numpy.ndarray, (..., n_chan_ref, filter_length)
        """
        n_col = col.shape[-1]
        # n_fft = 2 ** math.ceil(math.log2(2 * n_col - 1))
        n_fft = 2 * n_col
        pad_len = n_fft - 2 * n_col + 1
        pad_shape = col.shape[:-1] + (pad_len,)
        circ_col = np.concatenate(
            (col, np.zeros_like(col, shape=pad_shape), col[..., :0:-1]), axis=-1
        )
        self.Cforward = np.fft.rfft(circ_col, axis=-1)

        self._shape = col.shape + (col.shape[-1],)
        self.n_col = n_col
        self.n_fft = n_fft

    @property
    def shape(self):
        return self._shape

    @property
    def ndim(self):
        return len(self._shape)

    def __matmul__(self, lhs: np.ndarray) -> np.ndarray:
        if lhs.ndim == self.ndim:
            assert lhs.shape[-2] == self.shape[-1]
            Y = np.fft.rfft(lhs, n=self.n_fft, axis=-2)
            y = np.fft.irfft(Y * self.Cforward[..., None], n=self.n_fft, axis=-2)
            return y[..., : self.n_col, :]

        else:
            raise ValueError(
                "Dimension mismatch between operators with "
                f"shapes {self.shape} and {lhs.shape}"
            )


class BlockCirculantPreconditionerOperator:
    """
    Optimal circulant pre-conditioner operator for a symmetric topelitz matrix
    """

    def __init__(self, toeplitz_col: np.ndarray):
        """
        toeplitz_col: numpy.ndarray, (..., 2 * filter_length, n_channels, n_channels)
        """

        col_precond = optimal_nonsymmetric_circulant_precond_column(
            toeplitz_col, axis=-3
        )
        C = np.fft.rfft(col_precond, axis=-3)

        # complex pointwise inverse
        self.C = np.linalg.inv(C)
        self.n = col_precond.shape[-3]

        self._shape = col_precond.shape[:-1] + (
            self.n * self.C.shape[-2],
            self.n * self.C.shape[-1],
        )

    @property
    def shape(self):
        return self._shape

    @property
    def ndim(self):
        return len(self._shape)

    def __matmul__(self, lhs: np.ndarray) -> np.ndarray:

        assert lhs.shape[-2] == self.shape[-1], (
            "Dimension mismatch between operators with "
            f"shapes {self.shape} and {lhs.shape}"
        )

        lhs = lhs.reshape(lhs.shape[:-2] + (-1, self.C.shape[-1]) + lhs.shape[-1:])
        lhs = np.fft.rfft(lhs, n=self.n, axis=-3)
        prod = np.einsum("...rc,...cm->...rm", self.C, lhs)
        y = np.fft.irfft(prod, n=self.n, axis=-3)
        y = y.reshape(lhs.shape[:-3] + (-1,) + lhs.shape[-1:])

        return y


class BlockToeplitzOperator:
    """
    Operator implementing fast matrix multiplication by a block Toeplitz matrix.
    Each block of the matrix is a symmetric Toeplitz matrix.
    """

    def __init__(self, col: np.ndarray):
        """
        Parameters
        ----------
        col: numpy.ndarray, (..., 2 * n_toeplitz, n_block_rows, n_block_cols)
            The reduced representation  of the block Toeplitz matrix.
            The last dimension contains the concatenated first column and first row
            of the Toeplitz matrix.
            We can thus directly apply the FFT
        """
        n_col = col.shape[-3] // 2

        n_fft = 2 ** math.ceil(math.log2(2 * n_col))
        # n_fft = 2 * n_col

        pad_len = n_fft - 2 * n_col + 1
        pad_shape = col.shape[:-3] + (pad_len,) + col.shape[-2:]
        circ_col = np.concatenate(
            (
                col[..., :1, :, :],
                col[..., :-n_col:-1, :, :],
                np.zeros_like(col, shape=pad_shape),
                col[..., n_col - 1 : 0 : -1, :, :],
            ),
            axis=-3,
        )
        self.Cforward = np.fft.rfft(circ_col, axis=-3)

        self._n_block_rows = col.shape[-2]
        self._n_block_cols = col.shape[-1]
        self._n_toeplitz = n_col

        self.n_fft = circ_col.shape[-3]

        self._shape = col.shape[:-3] + (
            self._n_block_rows * n_col,
            self._n_block_cols * n_col,
        )
        self.n_fft = n_fft

    @property
    def shape(self):
        return self._shape

    @property
    def ndim(self):
        return len(self._shape)

    def __matmul__(self, lhs: np.ndarray) -> np.ndarray:
        assert lhs.shape[-2] == self.shape[-1], (
            "Dimension mismatch between operators with "
            f"shapes {self.shape} and {lhs.shape}"
        )
        # lhs.shape = (..., n_toeplitz, n_block_cols, n_lhs)
        lhs = lhs.reshape(lhs.shape[:-2] + (-1, self._n_block_cols) + lhs.shape[-1:])
        Y = np.fft.rfft(lhs, n=self.n_fft, axis=-3)
        prod = np.einsum("...rc,...cm->...rm", self.Cforward, Y)
        y = np.fft.irfft(prod, n=self.n_fft, axis=-3)

        # truncate output
        y = y[..., : self._n_toeplitz, :, :]

        # reshape as vector
        y = y.reshape(y.shape[:-3] + (-1,) + y.shape[-1:])

        return y


class IdentityOperator:
    def __matmul__(self, lhs: np.ndarray) -> np.ndarray:
        return lhs


def inner_prod(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    return np.einsum("...cd,...cd->...d", a.conj(), b)


def conjugate_gradient(
    A: Union[np.ndarray, BlockToeplitzOperator, SymmetricToeplitzOperator],
    b: np.ndarray,
    x: Optional[np.ndarray] = None,
    n_iter: Optional[int] = None,
    precond: Optional[
        Union[
            np.ndarray,
            BlockCirculantPreconditionerOperator,
            CirculantPreconditionerOperator,
        ]
    ] = None,
    verbose: Optional[bool] = False,
    x_true: Optional[np.ndarray] = None,
) -> np.ndarray:

    # assert hasattr(A, "__matmul__")

    add_axis = A.ndim - 1 == b.ndim
    if add_axis:
        b = b[..., None]

    assert len(A.shape) == len(b.shape) and A.shape[-1] == b.shape[-2]

    if x is None:
        x = np.zeros_like(b)

    if n_iter is None:
        n_iter = b.shape[-1]

    if precond is None:
        precond = IdentityOperator()

    r = b - A @ x
    z = precond @ r
    p = z

    rsold = inner_prod(r, z)

    if x_true is not None:
        if add_axis:
            x_true = x_true[..., None]
        errors = [np.mean(np.linalg.norm(x - x_true, axis=-2))]

    for epoch in range(n_iter):
        Ap = A @ p
        pAp = inner_prod(p, Ap)
        alpha = rsold / clamp(pAp, min=1e-6)
        x = x + alpha[..., None, :] * p
        if epoch + 1 % 5 == 0:
            r = b - A @ x
        else:
            r = r - alpha[..., None, :] * Ap
        z = precond @ r

        rsnew = inner_prod(r, z)
        if verbose:
            print(rsnew.mean())

        beta = rsnew / clamp(rsold, min=1e-6)
        p = z + beta[..., None, :] * p
        rsold = rsnew

        if x_true is not None:
            errors.append(np.mean(np.linalg.norm(x - x_true, axis=-2)))

    if add_axis:
        x = x[..., 0]

    if x_true is not None:
        return x, errors
    else:
        return x


def toeplitz_conjugate_gradient(
    acf: np.ndarray,
    xcorr: np.ndarray,
    x: Optional[np.ndarray] = None,
    n_iter: Optional[int] = None,
    x_true: Optional[np.ndarray] = None,
) -> np.ndarray:

    # prepare pre-conditioner
    precond = CirculantPreconditionerOperator(acf)

    # prepare forward operator
    forward = SymmetricToeplitzOperator(acf)

    # run conjugate gradient
    return conjugate_gradient(
        forward, xcorr, n_iter=n_iter, precond=precond, x=x, x_true=x_true
    )


def block_toeplitz_conjugate_gradient(
    acf: np.ndarray,
    xcorr: np.ndarray,
    x: Optional[np.ndarray] = None,
    n_iter: Optional[int] = None,
    x_true: Optional[np.ndarray] = None,
) -> np.ndarray:

    # prepare pre-conditioner
    precond = BlockCirculantPreconditionerOperator(acf)
    # precond = None

    # prepare forward operator
    forward = BlockToeplitzOperator(acf)

    # run conjugate gradient
    return conjugate_gradient(
        forward, xcorr, n_iter=n_iter, precond=precond, x=x, x_true=x_true
    )



# Copyright 2021 Robin Scheibler
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional, Tuple

import numpy as np

import torch

from .hungarian import linear_sum_assignment


def _remove_mean(x: torch.Tensor, dim: Optional[int] = -1) -> torch.Tensor:
    return x - x.mean(dim=dim, keepdim=True)


def _normalize(
    x: torch.Tensor, eps: Optional[float] = 1e-6, dim: Optional[int] = None
) -> torch.Tensor:
    x = x / torch.clamp(x.norm(dim=dim, keepdim=True), min=eps)
    return x


def _db_clamp_eps(db_max: float) -> float:
    e = 10.0 ** (-db_max / 10.0)
    eps = e / (1.0 + e)
    return eps


def _coherence_to_neg_sdr(
    coh: torch.Tensor, clamp_db: Optional[float] = 60.0
) -> torch.Tensor:

    if clamp_db is not None:
        # clamp within desired decibel range
        eps = _db_clamp_eps(clamp_db)
        coh = torch.clamp(coh, min=eps, max=1 - eps)

    ratio = (1 - coh) / coh

    # apply the SDR mapping
    return 10.0 * torch.log10(ratio)


def _solve_permutation(
    target_loss_matrix: torch.Tensor,
    *args,
    return_perm=False,
) -> Tuple[torch.Tensor]:
    """
    Solve the permutation in numpy for now
    """

    loss_mat = target_loss_matrix  # more consice name

    b_shape = loss_mat.shape[:-2]
    n_chan_ref, n_chan_est = loss_mat.shape[-2:]
    n_chan_out = min(n_chan_ref, n_chan_est)

    if n_chan_ref > n_chan_est:
        loss_mat = loss_mat.transpose(-2, -1)
        args = list(args)
        for i, arg in enumerate(args):
            args[i] = arg.transpose(-2, -1)

    loss_out = loss_mat.new_zeros(b_shape + (n_chan_out,))
    args_out = [arg.new_zeros(b_shape + (n_chan_out,)) for arg in args]

    p_opts = target_loss_matrix.new_zeros(b_shape + (n_chan_out,), dtype=torch.int64)
    for m in np.ndindex(b_shape):
        # linear sum assignment tries to *maximize* the sum, so we add a minus sign
        # because we are supposed to *minimize* losses
        dum, p_opt = _linear_sum_assignment_with_inf(loss_mat[m])
        loss_out[m] = loss_mat[m + (dum, p_opt)]
        for i, arg in enumerate(args):
            args_out[i][m] = arg[m + (dum, p_opt)]
        p_opts[m] = p_opt

    if return_perm:
        return (loss_out,) + tuple(args_out) + (p_opts,)
    else:
        if len(args_out) == 0:
            return loss_out
        else:
            return (loss_out,) + tuple(args_out)


def _linear_sum_assignment_with_inf(
    cost_matrix: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Solves the permutation problem efficiently via the linear sum
    assignment problem.

    https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.linear_sum_assignment.html
    This implementation was proposed by @louisabraham in
    https://github.com/scipy/scipy/issues/6900
    to handle infinite entries in the cost matrix.
    """

    if cost_matrix.numel() == 0:
        return torch.zeros(0, dtype=torch.int64), torch.zeros(0, dtype=torch.int64)

    min_inf = torch.isneginf(cost_matrix).any()
    max_inf = torch.isposinf(cost_matrix).any()
    if min_inf and max_inf:
        raise ValueError("matrix contains both inf and -inf")

    if min_inf or max_inf:
        cost_matrix = cost_matrix.clone()
        values = cost_matrix[~torch.isinf(cost_matrix)]
        m = values.min()
        M = values.max()
        n = min(cost_matrix.shape)
        # strictly positive constant even when added
        # to elements of the cost matrix
        positive = n * (M - m + torch.abs(M) + torch.abs(m) + 1)
        if max_inf:
            place_holder = (M + (n - 1) * (M - m)) + positive
        if min_inf:
            place_holder = (m + (n - 1) * (m - M)) - positive

        cost_matrix[torch.isinf(cost_matrix)] = place_holder

    return linear_sum_assignment(cost_matrix)



# Copyright 2021 Robin Scheibler
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import numpy as np


def hankel_view(x: np.ndarray, n_rows: int) -> np.ndarray:
    """return a view of x as a Hankel matrix"""
    n_cols = x.shape[-1] - n_rows + 1
    return np.lib.stride_tricks.as_strided(
        x, shape=x.shape[:-1] + (n_rows, n_cols), strides=x.strides + (x.strides[-1],)
    )


def toeplitz(x):
    """
    make symmetric toeplitz matrix
    """
    two_x = np.concatenate((x[..., :0:-1], x), axis=-1)
    H = hankel_view(two_x, n_rows=x.shape[-1])
    T = H[..., ::-1, :]
    return T


def block_toeplitz(x, n=None):
    """
    Create a block matrix where the blocks have a Toeplitz structure
    """
    top, rows, cols = x.shape[-3:]

    if n is None:
        n = top // 2

    out = np.zeros_like(x, shape=x.shape[:-3] + (n, rows, n, cols))

    for r in range(x.shape[-2]):
        for c in range(x.shape[-1]):
            vec = np.concatenate((x[..., -n + 1 :, r, c], x[..., :n, r, c]), axis=-1)
            T = hankel_view(vec, n)
            out[..., :, r, :, c] = T[..., ::-1, :]

    out = out.reshape(out.shape[:-4] + (n * rows, n * cols))

    return out



from typing import Optional, Tuple
import torch

from .metrics import sdr_loss, sdr, bss_eval_sources


def si_sdr_loss(
    est: torch.Tensor,
    ref: torch.Tensor,
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
    pairwise: Optional[bool] = False,
) -> torch.Tensor:
    return sdr_loss(
        est,
        ref,
        filter_length=1,
        zero_mean=zero_mean,
        clamp_db=clamp_db,
        pairwise=pairwise,
    )


def si_sdr_pit_loss(
    est: torch.Tensor,
    ref: torch.Tensor,
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
) -> torch.Tensor:
    return sdr(
        ref,
        est,
        filter_length=1,
        zero_mean=zero_mean,
        clamp_db=clamp_db,
        return_perm=False,
        change_sign=True,
    )


def pairwise_si_sdr_loss(
    est: torch.Tensor,
    ref: torch.Tensor,
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
) -> torch.Tensor:
    return si_sdr_loss(est, ref, zero_mean=zero_mean, clamp_db=clamp_db, pairwise=True)


def si_sdr(
    ref: torch.Tensor,
    est: torch.Tensor,
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
    return_perm: Optional[bool] = False,
    change_sign: Optional[bool] = False,
) -> Tuple[torch.Tensor, ...]:
    """
    Compute the scale-invariant signal-to-distortion ratio (SI-SDR) only.

    This function computes the SDR for all pairs of est/ref signals and finds the
    permutation maximizing the SDR.

    The order of ref/est follows the convention of bss_eval (i.e., ref first).

    Parameters
    ----------
    ref: torch.Tensor (..., n_channels_est, n_samples)
        The estimated signals
    est: torch.Tensor (..., n_channels_ref, n_samples)
        The groundtruth reference signals
    filter_length: int, optional
        The length of the distortion filter allowed (default: ``512``)
    use_cg_iter: int, optional
        If provided, an iterative method is used to solve for the distortion
        filter coefficients instead of direct Gaussian elimination.
        This can speed up the computation of the metrics in case the filters
        are long. Using a value of 10 here has been shown to provide
        good accuracy in most cases and is sufficient when using this
        loss to train neural separation networks.
    zero_mean: bool, optional
        When set to True, the mean of all signals is subtracted prior
        to computation of the metrics (default: ``False``)
    clamp_db: bool, optional
        If provided, the resulting metrics are clamped to be in the range ``[-clamp_db, clamp_db]``
    load_diag: float, optional
        If provided, this small value is added to the diagonal coefficients of
        the system metrics when solving for the filter coefficients.
        This can help stabilize the metric in the case where some of the reference
        signals may sometimes be zero
    return_perm: bool, optional
        If set to True, the optimal permutation of the estimated signals is
        also returned (default: ``False``)
    change_sign: bool, optional
        If set to True, the sign is flipped and the negative SDR is returned
        (default: ``False``)

    Returns
    -------
    cisdr: torch.Tensor, (..., n_channels_est)
        The SDR of the input signal
    perm: torch.Tensor, (..., n_channels_est), optional
        The index of the corresponding reference signal.
        Only returned if ``return_perm == True``
    """
    return sdr(
        ref,
        est,
        filter_length=1,
        use_cg_iter=None,
        zero_mean=zero_mean,
        clamp_db=clamp_db,
        return_perm=return_perm,
        change_sign=change_sign,
    )


def si_bss_eval_sources(
    ref: torch.Tensor,
    est: torch.Tensor,
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
    compute_permutation: Optional[bool] = True,
    load_diag: Optional[float] = None,
) -> torch.Tensor:
    """
    This function computes the SI-SDR, SI-SIR, and SI-SAR for the input
    reference and estimated signals.

    The order of ref/est follows the convention of bss_eval (i.e., ref first).

    Parameters
    ----------
    ref: torch.Tensor (..., n_channels_est, n_samples)
        The estimated signals
    est: torch.Tensor (..., n_channels_ref, n_samples)
        The groundtruth reference signals
    zero_mean: bool, optional
        When set to True, the mean of all signals is subtracted prior
        to computation of the metrics (default: ``False``)
    clamp_db: bool, optional
        If provided, the resulting metrics are clamped to be in the range ``[-clamp_db, clamp_db]``
    compute_permutation: bool, optional
        When this flag is true, the optimal permutation of the estimated signals
        to maximize the SIR is computed (default: ``True``)
    load_diag: float, optional
        If provided, this small value is added to the diagonal coefficients of
        the system metrics when solving for the filter coefficients.
        This can help stabilize the metric in the case where some of the reference
        signals may sometimes be zero

    Returns
    -------
    si-sdr: torch.Tensor, (..., n_channels_est)
        The scale-invariant signal-to-distortion-ratio (SDR)
    si-sir: torch.Tensor, (..., n_channels_est)
        The scale-invariant signal-to-interference-ratio (SIR)
    si-sar: torch.Tensor, (..., n_channels_est)
        The scale-invariant signal-to-interference-ratio (SIR)
    perm: torch.Tensor, (..., n_channels_est)
        The index of the corresponding reference signal (only when
        ``compute_permutation == True``)
    """

    return bss_eval_sources(
        ref,
        est,
        filter_length=1,
        use_cg_iter=None,
        zero_mean=zero_mean,
        clamp_db=clamp_db,
        compute_permutation=compute_permutation,
        load_diag=load_diag,
    )



# Copyright 2021 Robin Scheibler
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

r"""
This library implements the *bss_eval* metrics for the evaluation of blind
source separation (BSS) algorithms [1]_.  The implementation details and
benchmarks against other publicly available implementations are also available [2]_.

Quick start
-----------

The package can be installed from pip. At a minimum, it requires to have ``numpy``/``scipy`` installed.
To work with torch tensors, ``pytorch`` should also be installed obviously.

..  code-block:: shell

    pip install fast-bss-eval

Assuming you have multichannel signals for the estimated and reference sources
stored in wav format files named ``my_estimate_file.wav`` and
``my_reference_file.wav``, respectively, you can quickly evaluate the bss_eval
metrics as follows with the function :py:func:`bss_eval_sources` and :py:func:`sdr`.

.. code-block:: python

    from scipy.io import wavfile
    import fast_bss_eval

    # open the files, we assume the sampling rate is known
    # to be the same
    fs, ref = wavfile.read("my_reference_file.wav")
    _, est = wavfile.read("my_estimate_file.wav")

    # compute all the metrics
    sdr, sir, sar, perm = fast_bss_eval.bss_eval_sources(ref.T, est.T)

    # compute the SDR only
    sdr = fast_bss_eval.sdr(ref.T, est.T)

You can use the :py:func:`sdr_pit_loss` and :py:func:`sdr_loss` to train separation networks with
`pytorch`_ and
`torchaudio <https://github.com/pytorch/audio>`_.
The difference with the regular functions is as follows.

- reverse order of ``est`` and ``ref`` to follow pytorch convention
- sign of SDR is changed to act as a loss function

For permutation invariant training (PIT), use :py:func:`sdr_pit_loss`.  Use
:py:func:`sdr_loss` if the assignement of estimates to references is known, or
to compute the pairwise metrics.

.. code-block:: python

    import torchaudio

    # open the files, we assume the sampling rate is known
    # to be the same
    ref, fs = torchaudio.load("my_reference_file.wav")
    est, _ = torchaudio.load("my_estimate_file.wav")

    # PIT loss for permutation invariant training
    neg_sdr = fast_bss_eval.sdr_pit_loss(est, ref)

    # when the assignemnt of estimates to references is known
    neg_sdr = fast_bss_eval.sdr_loss(est, ref)

To summarize, there are four functions implemented, :func:`bss_eval_sources`,
:func:`sdr`, and :func:`sdr_pit_loss`, and :func:`sdr_loss`.
The **scale-invariant** [3]_ versions of all these functions are available in functions pre-fixed by ``si_``, i.e.,

* :py:func:`si_bss_eval_sources`
* :py:func:`si_sdr`
* :py:func:`si_sdr_pit_loss`
* :py:func:`si_sdr_loss`.

Alternatively, the regular functions with keyword argument ``filter_length=1`` can be used.

.. note::

    If you use this package in your own research, please cite `our paper <https://arxiv.org/abs/2110.06440>`_ describing it [4]_.

    .. code-block:: bibtex

        @misc{scheibler_sdr_2021,
          title={SDR --- Medium Rare with Fast Computations},
          author={Robin Scheibler},
          year={2021},
          eprint={2110.06440},
          archivePrefix={arXiv},
          primaryClass={eess.AS}
        }

Benchmark
---------

Average execution time on a dataset of multichannel signals with 2, 3, and 4
channels.  We compare ``fast_bss_eval`` to `mir_eval
<https://github.com/craffel/mir_eval>`_ and `sigsep
<https://github.com/sigsep/bsseval>`_.  We further benchmark ``fast_bss_eval``
using ``numpy``/``pytorch`` with double and single precision floating-point
arithmetic (fp64/fp32), and using ``solve`` function or an :ref:`iterative solver<Iterative solver for the disortion filters>` (CGD10).

The benchmark was run on a Linux workstation with an Intel Xeon Gold 6230 CPU @
2.10 GHz with 8 cores, an NVIDIA Tesla V100 GPU, and 64 GB of RAM.
We test single-threaded (1CPU) and 8 core performance (8CPU) on the CPU, and on the GPU.
We observe that for ``numpy`` there is not so much gain of using multiple cores.
It may be more efficient to run multiple samples in parallel single-threaded
processes to fully exploit multi-core architectures.


.. figure:: figures/channels_vs_runtime.png
    :alt: runtime benchmark result

bss_eval metrics theory
------------------------

:math:`\newcommand{\vs}{\boldsymbol{s}}\newcommand{\vshat}{\hat{\vs}}\newcommand{\vb}{\boldsymbol{b}}`
We consider the case where we have :math:`M` estimated signals :math:`\vshat_m`.
Each contains a convolutional mixture of :math:`K` reference signals
:math:`\vs_k` and an additive noise component :math:`\vb_m`,

.. math::

    \newcommand{\vh}{\boldsymbol{h}}
    \newcommand{\vg}{\boldsymbol{g}}

    \vshat_m = \sum\nolimits_{k}\vh_{mk} \star \vs_k + \vb_m,
    \quad m=1,\ldots,M,

where :math:`\vshat_m`, :math:`\vs_k`, and :math:`\vb_m` are all real vectors
of length :math:`T`.  The length of the impulse responses :math:`\vh_{mk}` is
assumed to be short compared to :math:`T`.  For simplicity, the convolution
operation here includes truncation to size :math:`T`.  In most cases, the
number of estimates and references is the same, i.e. :math:`M=K`.  We keep them
distinct for generality.

.. figure:: figures/bss_eval_metrics_projections.png
    :scale: 50 %
    :alt: figure describing how bss_eval metrics are derived

bss_eval decomposes the estimated signal into three *orthogonal* components corresponding to the
target, interference, and noise, as shown in the illustration above.

.. math::
    \newcommand{\ve}{\boldsymbol{e}}
    \newcommand{\mP}{\boldsymbol{P}}
    \newcommand{\mA}{\boldsymbol{A}}
    \newcommand{\starget}{\vs^{\text{target}}}
    \newcommand{\einterf}{\ve^{\text{interf}}}
    \newcommand{\eartif}{\ve^{\text{artif}}}

    \starget_{km} = \mP_{k} \vshat_m, \quad
    \einterf_{km} = \mP \vshat_m - \mP_k \vshat_m, \quad
    \eartif_{km} = \vshat_m - \mP \vshat_m.

Rather than projecting directly on the signals themselves, bss_eval allows for
a short distortion filter of length :math:`L` to be applied.
This corresponds to projection onto the subspace spanned by the :math:`L`
shifts of the signals.
Matrices :math:`\mP_k` and :math:`\mP` are orthogonal projection matrices onto
the subspace spaned by the :math:`L` shifts of :math:`\vs_k` and of all
references, respectively.  Let :math:`\mA_k \in \mathbb{R}^{(T + L - 1) \times
L}` contain the :math:`L` shifts of :math:`\vs_k` in its columns and :math:`\mA
= [\mA_1,\ldots,\mA_K]`, then

.. math::

    \mP_k = \mA_k(\mA_k^\top \mA_k)^{-1}\mA_k^\top,\quad
    \mP = \mA(\mA^\top \mA)^{-1}\mA^\top.

Then, SDR, SIR, and SAR, in decibels, are defined as follows,

.. math::

    \text{SDR}_{km} & = 10\log_{10} \frac{\| \starget_{km} \|^2}{\|\einterf_{km} + \eartif_{km} \|^2}, \\
    \text{SIR}_{km} & = 10\log_{10} \frac{\| \starget_{km} \|^2}{\| \einterf_{km} \|^2}, \\
    \text{SAR}_{km} & = 10\log_{10} \frac{\| \starget_{km} + \einterf_{km} \|^2}{\| \eartif_{m} \|^2}.

Finally, assuming for simplicity that :math:`K=M`, the permutation of the estimated
sources :math:`\pi\,:\,\{1,\ldots,K\} \to \{1,\ldots,K\}` that maximizes :math:`\sum_k \text{SIR}_{k\,\pi(k)}`
is chosen.
Once we have the matrix of SIR values, we can use the `linear_sum_assignement
<https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.linear_sum_assignment.html>`_
function from scipy to find this optimal permutation efficiently.

The scale-invariant SDR is equivalent to using a distortion filter length of 1
[3]_.  The use of longer filters has been shown to be beneficial when training
neural networks to assist beamforming algorithms [2]_ [4]_.

Iterative solver for the disortion filters
------------------------------------------

.. note::

    **tl;dr** If you need more speed, add the keyword argument ``use_cg_iter=10`` to any method call.

As described above, computing the bss_eval metrics requires to solve the systems

.. math::

    (\mA_k^\top \mA_k) \vh_{km} & = \mA_k^\top \vshat_m, \\
    (\mA^\top \mA) \vg_{m} & = \mA^\top \vshat_m, \\

for all :math:`k` and :math:`m`.  These system matrices are large and solving
directly, e.g., using ``numpy.linalg.solve``, can be expensive.  However, these
matrices are also structured (`Toeplitz
<https://en.wikipedia.org/wiki/Toeplitz_matrix>`_ and block-Toeplitz).  In this
case, we can apply a much more efficient iterative algorithm based on
`preconditioned conjugate gradient descent
<https://en.wikipedia.org/wiki/Conjugate_gradient_method>`_ [5]_.

The default behavior of ``fast_bss_eval`` is to use the direct solver (via the
``solve`` function of ``numpy`` or ``pytorch``) to ensure consistency with
the output of other packages.
However, if speed becomes an issue, it is possible to switch to the iterative
solver by adding the option ``use_cg_iter`` (USE Conjugate Gradient ITERations) to
any of the methods in ``fast_bss_eval``.
We found that using 10 iterations is a good trade-off between speed and accuracy,
i.e., ``use_cg_iter=10``.

Author
------

This package was written by `Robin Scheibler <http://www.robinscheibler.com>`_.

License
-------

2021 (c) Robin Scheibler, LINE Corporation

The source code for this package is released under `MIT License
<https://opensource.org/licenses/MIT>`_ and available on `github
<https://github.com/fakufaku/fast_bss_eval>`_.

Bibliography
------------

.. [1] E. Vincent, R. Gribonval, and C. Fevotte, *Performance measurement in
    blind audio source separation*, IEEE Trans. Audio Speech Lang. Process., vol.
    14, no. 4, pp. 1462–1469, Jun. 2006.

.. [2] R\. Scheibler, *SDR - Medium Rare with Fast Computations*, arXiv:2110.06440, 2021. `pdf <https://arxiv.org/abs/2110.06440>`_

.. [3] J. Le Roux, S. Wisdom, H. Erdogan, and J. R. Hershey, *SDR - Half-baked or Well Done?*,
    In Proc. IEEE ICASSP, Brighton, UK, May 2019.

.. [4] C. Boeddeker et al., *Convolutive transfer function invariant SDR training criteria
    for multi-channel reverberant speech separation*, in Proc. IEEE ICASSP, Toronto,
    CA, Jun. 2021, pp. 8428–8432.

.. [5] R. H. Chan and M. K. Ng, *Conjugate gradient methods for Toeplitz
    systems*, SIAM Review, vol. 38, no. 3, pp. 427–482, Sep. 1996.
"""
from typing import Optional, Union, Tuple

import numpy as np

try:
    import torch as pt

    has_torch = True
    from . import torch as torch
    from .torch import sdr_pit_loss, si_sdr_pit_loss
except ImportError:
    has_torch = False

    # dummy pytorch module
    class pt:
        class Tensor:
            def __init__(self):
                pass

    # dummy torch submodule
    class torch:
        bss_eval_sources = None
        sdr = None
        sdr_loss = None


from . import numpy as numpy


def _dispatch_backend(f_numpy, f_torch, *args, **kwargs):
    if has_torch and isinstance(args[0], pt.Tensor):
        return f_torch(*args, **kwargs)
    else:
        return f_numpy(*args, **kwargs)


def bss_eval_sources(
    ref: Union[np.ndarray, pt.Tensor],
    est: Union[np.ndarray, pt.Tensor],
    filter_length: Optional[int] = 512,
    use_cg_iter: Optional[int] = None,
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
    compute_permutation: Optional[bool] = True,
    load_diag: Optional[float] = None,
) -> Union[Tuple[np.ndarray, ...], Tuple[pt.Tensor, ...]]:
    """
    This function computes the SDR, SIR, and SAR for the input reference and estimated
    signals.

    The order of ``ref``/``est`` follows the convention of bss_eval (i.e., ref first).

    Parameters
    ----------
    ref:
        The estimated signals, ``shape == (..., n_channels_est, n_samples)``
    est:
        The groundtruth reference signals, ``shape ==  (..., n_channels_ref, n_samples)``
    filter_length:
        The length of the distortion filter allowed (default: ``512``)
    use_cg_iter:
        If provided, an iterative method is used to solve for the distortion
        filter coefficients instead of direct Gaussian elimination.
        This can speed up the computation of the metrics in case the filters
        are long. Using a value of 10 here has been shown to provide
        good accuracy in most cases and is sufficient when using this
        loss to train neural separation networks.
    zero_mean:
        When set to True, the mean of all signals is subtracted prior
        to computation of the metrics (default: ``False``)
    clamp_db:
        If provided, the resulting metrics are clamped to be in the range ``[-clamp_db, clamp_db]``
    compute_permutation:
        When this flag is true, the optimal permutation of the estimated signals
        to maximize the SIR is computed (default: ``True``)
    load_diag:
        If provided, this small value is added to the diagonal coefficients of
        the system metrics when solving for the filter coefficients.
        This can help stabilize the metric in the case where some of the reference
        signals may sometimes be zero

    Returns
    -------
    sdr:
        The signal-to-distortion-ratio (SDR), ``shape == (..., n_channels_est)``
    sir:
        The signal-to-interference-ratio (SIR), ``shape == (..., n_channels_est)``
    sar:
        The signal-to-interference-ratio (SIR), ``shape == (..., n_channels_est)``
    perm:
        The index of the corresponding reference signal (only when
        ``compute_permutation == True``), ``shape == (..., n_channels_est)``
    """

    return _dispatch_backend(
        numpy.bss_eval_sources,
        torch.bss_eval_sources,
        ref,
        est,
        filter_length=filter_length,
        use_cg_iter=use_cg_iter,
        zero_mean=zero_mean,
        clamp_db=clamp_db,
        compute_permutation=compute_permutation,
        load_diag=load_diag,
    )


def sdr(
    ref: Union[np.ndarray, pt.Tensor],
    est: Union[np.ndarray, pt.Tensor],
    filter_length: Optional[int] = 512,
    use_cg_iter: Optional[int] = None,
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
    load_diag: Optional[float] = None,
    return_perm: Optional[bool] = False,
    change_sign: Optional[bool] = False,
) -> Union[np.ndarray, pt.Tensor]:
    """
    Compute the signal-to-distortion ratio (SDR) only.

    This function computes the SDR for all pairs of est/ref signals and finds the
    permutation maximizing the sum of all SDRs. This is unlike :py:func:`bss_eval_sources`
    that uses the SIR.

    The order of ``ref``/``est`` follows the convention of bss_eval (i.e., ref first).

    Parameters
    ----------
    ref:
        The estimated signals, ``shape == (..., n_channels_est, n_samples)``
    est:
        The groundtruth reference signals, ``shape == (..., n_channels_ref, n_samples)``
    filter_length:
        The length of the distortion filter allowed (default: ``512``)
    use_cg_iter:
        If provided, an iterative method is used to solve for the distortion
        filter coefficients instead of direct Gaussian elimination.
        This can speed up the computation of the metrics in case the filters
        are long. Using a value of 10 here has been shown to provide
        good accuracy in most cases and is sufficient when using this
        loss to train neural separation networks.
    zero_mean:
        When set to True, the mean of all signals is subtracted prior
        to computation of the metrics (default: ``False``)
    clamp_db:
        If provided, the resulting metrics are clamped to be in the range ``[-clamp_db, clamp_db]``
    load_diag:
        If provided, this small value is added to the diagonal coefficients of
        the system metrics when solving for the filter coefficients.
        This can help stabilize the metric in the case where some of the reference
        signals may sometimes be zero
    return_perm:
        If set to True, the optimal permutation of the estimated signals is
        also returned (default: ``False``)
    change_sign:
        If set to True, the sign is flipped and the negative SDR is returned
        (default: ``False``)

    Returns
    -------
    sdr:
        The SDR of the input signal ``shape == (..., n_channels_est)``
    perm:
        The index of the corresponding reference signal ``shape == (..., n_channels_est)``.
        Only returned if ``return_perm == True``
    """
    return _dispatch_backend(
        numpy.sdr,
        torch.sdr,
        ref,
        est,
        filter_length=filter_length,
        use_cg_iter=use_cg_iter,
        zero_mean=zero_mean,
        clamp_db=clamp_db,
        load_diag=load_diag,
        return_perm=return_perm,
        change_sign=change_sign,
    )


def sdr_loss(
    est: Union[np.ndarray, pt.Tensor],
    ref: Union[np.ndarray, pt.Tensor],
    filter_length: Optional[int] = 512,
    use_cg_iter: Optional[int] = None,
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
    load_diag: Optional[float] = None,
    pairwise: Optional[bool] = False,
) -> Union[np.ndarray, pt.Tensor]:
    """
    Computes the negative signal-to-distortion ratio (SDR).

    This function is almost the same as :py:func:`fast_bss_eval.sdr` except for

    * negative sign to make it a *loss* function
    * ``est``/``ref`` arguments follow the convention of pytorch loss functions (i.e., ``est`` first).

    Parameters
    ----------
    est:
        The estimated signals ``shape == (..., n_channels_est, n_samples)``
    ref:
        The groundtruth reference signals ``shape == (..., n_channels_ref, n_samples)``
    filter_length:
        The length of the distortion filter allowed (default: ``512``)
    use_cg_iter:
        If provided, an iterative method is used to solve for the distortion
        filter coefficients instead of direct Gaussian elimination.
        This can speed up the computation of the metrics in case the filters
        are long. Using a value of 10 here has been shown to provide
        good accuracy in most cases and is sufficient when using this
        loss to train neural separation networks.
    zero_mean:
        When set to True, the mean of all signals is subtracted prior
        to computation of the metrics (default: ``False``)
    clamp_db:
        If provided, the resulting metrics are clamped to be in the range ``[-clamp_db, clamp_db]``
    load_diag:
        If provided, this small value is added to the diagonal coefficients of
        the system metrics when solving for the filter coefficients.
        This can help stabilize the metric in the case where some of the reference
        signals may sometimes be zero
    pairwise:
        If set to True, the metrics are computed for every est/ref signals pair (default: ``False``).
        When set to False, it is expected that ``n_channels_ref == n_channels_est``.

    Returns
    -------
    :
        The negative SDR of the input signal. The returned tensor has

        * ``shape == (..., n_channels_est)`` if ``pairwise == False``
        * ``shape == (..., n_channels_ref, n_channels_est)`` if ``pairwise == True``.
    """
    return _dispatch_backend(
        numpy.sdr_loss,
        torch.sdr_loss,
        est,
        ref,
        filter_length=filter_length,
        use_cg_iter=use_cg_iter,
        zero_mean=zero_mean,
        clamp_db=clamp_db,
        pairwise=pairwise,
        load_diag=load_diag,
    )


def sdr_pit_loss(
    est: Union[np.ndarray, pt.Tensor],
    ref: Union[np.ndarray, pt.Tensor],
    filter_length: Optional[int] = 512,
    use_cg_iter: Optional[int] = None,
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
    load_diag: Optional[float] = None,
) -> Union[np.ndarray, pt.Tensor]:
    """
    Computes the negative signal-to-distortion ratio (SDR) with the PIT loss
    (permutation invariant training).  This means the permutation of sources
    minimizing the loss is chosen.

    This function is similar to :py:func:`fast_bss_eval.sdr` and :py:func:`fast_bss_eval.sdr_loss`, but with,

    * negative sign to make it a *loss* function
    * ``est``/``ref`` arguments follow the convention of pytorch loss functions (i.e., ``est`` first).
    * pairwise and permutation solving is used (a.k.a. PIT, permutation invariant training)

    Parameters
    ----------
    est:
        The estimated signals ``shape == (..., n_channels_est, n_samples)``
    ref:
        The groundtruth reference signals ``shape == (..., n_channels_ref, n_samples)``
    filter_length:
        The length of the distortion filter allowed (default: ``512``)
    use_cg_iter:
        If provided, an iterative method is used to solve for the distortion
        filter coefficients instead of direct Gaussian elimination.
        This can speed up the computation of the metrics in case the filters
        are long. Using a value of 10 here has been shown to provide
        good accuracy in most cases and is sufficient when using this
        loss to train neural separation networks.
    zero_mean:
        When set to True, the mean of all signals is subtracted prior
        to computation of the metrics (default: ``False``)
    clamp_db:
        If provided, the resulting metrics are clamped to be in the range ``[-clamp_db, clamp_db]``
    load_diag:
        If provided, this small value is added to the diagonal coefficients of
        the system metrics when solving for the filter coefficients.
        This can help stabilize the metric in the case where some of the reference
        signals may sometimes be zero

    Returns
    -------
    :
        The negative SDR of the input signal after solving for the optimal permutation.
        This is equivalent to using ``sdr_loss`` with ``pairwise=True`` with a PIT loss.
        The returned tensor has

        * ``shape == (..., n_channels_est)`` if ``pairwise == False``
    """
    return _dispatch_backend(
        numpy.sdr_pit_loss,
        torch.sdr_pit_loss,
        est,
        ref,
        filter_length=filter_length,
        use_cg_iter=use_cg_iter,
        zero_mean=zero_mean,
        clamp_db=clamp_db,
        load_diag=load_diag,
    )


def si_bss_eval_sources(
    ref: Union[np.ndarray, pt.Tensor],
    est: Union[np.ndarray, pt.Tensor],
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
    compute_permutation: Optional[bool] = True,
    load_diag: Optional[float] = None,
) -> Union[Tuple[np.ndarray, ...], Tuple[pt.Tensor, ...]]:
    """
    This function computes the scale-invariant metrics, SI-SDR, SI-SIR, and
    SI-SAR, for the input reference and estimated signals.

    The order of ``ref``/``est`` follows the convention of bss_eval (i.e., ref first).

    Parameters
    ----------
    ref:
        The estimated signals, ``shape == (..., n_channels_est, n_samples)``
    est:
        The groundtruth reference signals, ``shape ==  (..., n_channels_ref, n_samples)``
    zero_mean:
        When set to True, the mean of all signals is subtracted prior
        to computation of the metrics (default: ``False``)
    clamp_db:
        If provided, the resulting metrics are clamped to be in the range ``[-clamp_db, clamp_db]``
    compute_permutation:
        When this flag is true, the optimal permutation of the estimated signals
        to maximize the SIR is computed (default: ``True``)
    load_diag:
        If provided, this small value is added to the diagonal coefficients of
        the system metrics when solving for the filter coefficients.
        This can help stabilize the metric in the case where some of the reference
        signals may sometimes be zero

    Returns
    -------
    si-sdr:
        The signal-to-distortion-ratio (SDR), ``shape == (..., n_channels_est)``
    si-sir:
        The signal-to-interference-ratio (SIR), ``shape == (..., n_channels_est)``
    si-sar:
        The signal-to-interference-ratio (SIR), ``shape == (..., n_channels_est)``
    perm:
        The index of the corresponding reference signal (only when
        ``compute_permutation == True``), ``shape == (..., n_channels_est)``
    """

    return _dispatch_backend(
        numpy.si_bss_eval_sources,
        torch.si_bss_eval_sources,
        ref,
        est,
        zero_mean=zero_mean,
        clamp_db=clamp_db,
        compute_permutation=compute_permutation,
        load_diag=load_diag,
    )


def si_sdr(
    ref: Union[np.ndarray, pt.Tensor],
    est: Union[np.ndarray, pt.Tensor],
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
    return_perm: Optional[bool] = False,
    change_sign: Optional[bool] = False,
) -> Union[np.ndarray, pt.Tensor]:
    """
    Compute the scale-invariant signal-to-distortion ratio (SI-SDR) only.

    This function computes the SDR for all pairs of est/ref signals and finds the
    permutation maximizing the sum of all SDRs. This is unlike :py:func:`fast_bss_eval.bss_eval_sources`
    that uses the SIR.

    The order of ``ref``/``est`` follows the convention of bss_eval (i.e., ref first).

    Parameters
    ----------
    ref:
        The estimated signals, ``shape == (..., n_channels_est, n_samples)``
    est:
        The groundtruth reference signals, ``shape == (..., n_channels_ref, n_samples)``
    filter_length:
        The length of the distortion filter allowed (default: ``512``)
    use_cg_iter:
        If provided, an iterative method is used to solve for the distortion
        filter coefficients instead of direct Gaussian elimination.
        This can speed up the computation of the metrics in case the filters
        are long. Using a value of 10 here has been shown to provide
        good accuracy in most cases and is sufficient when using this
        loss to train neural separation networks.
    zero_mean:
        When set to True, the mean of all signals is subtracted prior
        to computation of the metrics (default: ``False``)
    clamp_db:
        If provided, the resulting metrics are clamped to be in the range ``[-clamp_db, clamp_db]``
    return_perm:
        If set to True, the optimal permutation of the estimated signals is
        also returned (default: ``False``)
    change_sign:
        If set to True, the sign is flipped and the negative SDR is returned
        (default: ``False``)

    Returns
    -------
    si-sdr:
        The SDR of the input signal ``shape == (..., n_channels_est)``
    perm:
        The index of the corresponding reference signal ``shape == (..., n_channels_est)``.
        Only returned if ``return_perm == True``
    """
    return _dispatch_backend(
        numpy.si_sdr,
        torch.si_sdr,
        ref,
        est,
        zero_mean=zero_mean,
        clamp_db=clamp_db,
        return_perm=return_perm,
        change_sign=change_sign,
    )


def si_sdr_loss(
    est: Union[np.ndarray, pt.Tensor],
    ref: Union[np.ndarray, pt.Tensor],
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
    pairwise: Optional[bool] = False,
) -> Union[np.ndarray, pt.Tensor]:
    """
    Computes the negative scale-invariant signal-to-distortion ratio (SI-SDR).

    This function is almost the same as :py:func:`fast_bss_eval.sdr` except for

    * negative sign to make it a *loss* function
    * ``est``/``ref`` arguments follow the convention of pytorch loss functions (i.e., ``est`` first).

    Parameters
    ----------
    est:
        The estimated signals ``shape == (..., n_channels_est, n_samples)``
    ref:
        The groundtruth reference signals ``shape == (..., n_channels_ref, n_samples)``
    zero_mean:
        When set to True, the mean of all signals is subtracted prior
        to computation of the metrics (default: ``False``)
    clamp_db:
        If provided, the resulting metrics are clamped to be in the range ``[-clamp_db, clamp_db]``
    pairwise:
        If set to True, the metrics are computed for every est/ref signals pair (default: ``False``).
        When set to False, it is expected that ``n_channels_ref == n_channels_est``.

    Returns
    -------
    :
        The negative SI-SDR of the input signal. The returned tensor has

        * ``shape == (..., n_channels_est)`` if ``pairwise == False``
        * ``shape == (..., n_channels_ref, n_channels_est)`` if ``pairwise == True``.
    """
    return _dispatch_backend(
        numpy.si_sdr_loss,
        torch.si_sdr_loss,
        est,
        ref,
        zero_mean=zero_mean,
        clamp_db=clamp_db,
        pairwise=pairwise,
    )


def si_sdr_pit_loss(
    est: Union[np.ndarray, pt.Tensor],
    ref: Union[np.ndarray, pt.Tensor],
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
) -> Union[np.ndarray, pt.Tensor]:
    """
    Computes the negative scale-invariant signal-to-distortion ratio (SDR) with
    the PIT loss (permutation invariant training).  This means the permutation
    of sources minimizing the loss is chosen.

    This function is similar to :py:func:`fast_bss_eval.sdr` and :py:func:`fast_bss_eval.sdr_loss`, but with,

    * negative sign to make it a *loss* function
    * ``est``/``ref`` arguments follow the convention of pytorch loss functions (i.e., ``est`` first).
    * pairwise and permutation solving is used (a.k.a. PIT, permutation invariant training)

    Parameters
    ----------
    est:
        The estimated signals ``shape == (..., n_channels_est, n_samples)``
    ref:
        The groundtruth reference signals ``shape == (..., n_channels_ref, n_samples)``
    zero_mean:
        When set to True, the mean of all signals is subtracted prior
        to computation of the metrics (default: ``False``)
    clamp_db:
        If provided, the resulting metrics are clamped to be in the range ``[-clamp_db, clamp_db]``

    Returns
    -------
    :
        The negative SDR of the input signal after solving for the optimal permutation.
        This is equivalent to using ``sdr_loss`` with ``pairwise=True`` with a PIT loss.
        The returned tensor has

        * ``shape == (..., n_channels_est)`` if ``pairwise == False``
    """
    return _dispatch_backend(
        numpy.si_sdr_pit_loss,
        torch.si_sdr_pit_loss,
        est,
        ref,
        zero_mean=zero_mean,
        clamp_db=clamp_db,
    )



# Copyright 2021 Robin Scheibler
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import math
from typing import Optional, Tuple

import torch

from .cgd import block_toeplitz_conjugate_gradient, toeplitz_conjugate_gradient
from .helpers import (
    _coherence_to_neg_sdr,
    _normalize,
    _remove_mean,
    _solve_permutation,
)
from .linalg import toeplitz, block_toeplitz


def square_cosine_metrics_length_one_filter(
    ref: torch.Tensor,
    est: torch.Tensor,
    zero_mean: Optional[bool] = False,
    pairwise: Optional[bool] = True,
    load_diag: Optional[float] = None,
    with_coh_sar: Optional[bool] = True,
) -> Tuple[torch.Tensor, ...]:
    """
    Special case of computing cosine metrics with filter length of 1
    """

    if zero_mean:
        ref = _remove_mean(ref, dim=-1)
        est = _remove_mean(est, dim=-1)

    ref = _normalize(ref, dim=-1)
    est = _normalize(est, dim=-1)

    if pairwise or with_coh_sar:
        xcorr = torch.einsum("...cn,...dn->...cd", ref, est)

    if pairwise:
        coh_sdr = xcorr
    else:
        coh_sdr = torch.einsum("...n,...n->...", ref, est)
    coh_sdr = torch.square(coh_sdr)

    if with_coh_sar:
        acm = torch.einsum("...cn,...dn->...cd", ref, ref)
        if load_diag is not None:
            acm = acm + torch.eye(acm.shape[-1]) * load_diag
        sol = torch.linalg.solve(acm, xcorr)
        coh_sar = torch.einsum("...lc,...lc->...c", xcorr, sol)

        if pairwise:
            coh_sdr, coh_sar = torch.broadcast_tensors(coh_sdr, coh_sar[..., None, :])

        return coh_sdr, coh_sar

    else:
        return coh_sdr


def compute_stats(
    x: torch.Tensor,
    y: torch.Tensor,
    length: Optional[int] = None,
    pairwise: Optional[bool] = True,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Compute the auto correlation function of x and its cross-correlation with y.
    This function is specialized for when only the SDR is needed.
    In this case, only the auto-correlation of each source is required.

    Parameters
    ----------
    x: torch.Tensor, (..., n_chan_x, n_samples)
        Usually the reference signals
    y: torch.Tensor, (..., n_chan_y, n_samples)
        Usually the estimated signals
    length: int, optional
        The length at which to truncate the statistics
    pairwise: bool, optional
        When this flag is true, statistics are computed for all
        pairs of source/estimate. This only affects the cross-correlation.

    Returns
    -------
    acf: (..., n_chan_x, length)
        The auto-correlation functions of x
    xcorr: torch.Tensor
        The cross-correlation functions of x and y.
        The shape is (..., n_chan_x, length, n_chan_y) when pairwise is True, and
        (..., n_chan_x, length) when pairwise is False.
    """
    if length is None:
        length = x.shape[-1]

    max_len = max(x.shape[-1], y.shape[-1])
    n_fft = 2 ** math.ceil(math.log2(x.shape[-1] + max_len - 1))

    X = torch.fft.rfft(x, n=n_fft, dim=-1)
    Y = torch.fft.rfft(y, n=n_fft, dim=-1)

    # autocorrelation function
    acf = torch.fft.irfft(X.real**2 + X.imag**2, n=n_fft)

    # cross-correlation
    if pairwise:
        XY = torch.einsum("...cn,...dn->...cnd", X.conj(), Y)
        xcorr = torch.fft.irfft(XY, n=n_fft, dim=-2)
        return acf[..., :length], xcorr[..., :length, :]

    else:
        XY = X.conj() * Y
        xcorr = torch.fft.irfft(XY, n=n_fft, dim=-1)
        return acf[..., :length], xcorr[..., :length]


def compute_stats_2(
    x: torch.Tensor,
    y: torch.Tensor,
    length: Optional[int] = None,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Compute the auto correlation function of x and its cross-correlation with y.
    This function computes the full auto-correlation matrix for all reference signals.
    This is only needed when both the SDR and the SIR have to be computed.

    Parameters
    ----------
    x: torch.Tensor, (..., n_chan_x, n_samples)
        Usually the reference signals
    y: torch.Tensor, (..., n_chan_y, n_samples)
        Usually the estimated signals
    length: int, optional
        The length at which to truncate the statistics

    Returns
    -------
    acf: (..., length, n_chan_x, n_chan_y)
        The auto-correlation functions of x
    xcorr: torch.Tensor, (..., length, n_chan_x, n_chan_y)
        The cross-correlation functions of y
    """
    if length is None:
        length = x.shape[-1]

    max_len = max(x.shape[-1], y.shape[-1])
    n_fft = 2 ** math.ceil(math.log2(x.shape[-1] + max_len - 1))

    X = torch.fft.rfft(x, n=n_fft, dim=-1)
    Y = torch.fft.rfft(y, n=n_fft, dim=-1)

    # autocorrelation function
    prod = torch.einsum("...cn,...dn->...ncd", X, X.conj())
    acf = torch.fft.irfft(prod, n=n_fft, dim=-3)
    acf = torch.cat([acf[..., :length, :, :], acf[..., -length:, :, :]], dim=-3)

    # cross-correlation
    XY = torch.einsum("...cn,...dn->...ncd", X.conj(), Y)
    xcorr = torch.fft.irfft(XY, n=n_fft, dim=-3)
    xcorr = xcorr[..., :length, :, :]

    return acf, xcorr


def sdr_loss(
    est: torch.Tensor,
    ref: torch.Tensor,
    filter_length: Optional[int] = 512,
    use_cg_iter: Optional[int] = None,
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
    load_diag: Optional[float] = None,
    pairwise: Optional[bool] = False,
) -> torch.Tensor:
    """
    Compute the negative signal-to-distortion ratio (SDR).

    The order of est/ref follows the convention of pytorch loss functions
    (i.e., est first).

    Parameters
    ----------
    est: torch.Tensor (..., n_channels_est, n_samples)
        The estimated signals
    ref: torch.Tensor (..., n_channels_ref, n_samples)
        The groundtruth reference signals
    filter_length: int, optional
        The length of the distortion filter allowed (default: ``512``)
    use_cg_iter: int, optional
        If provided, an iterative method is used to solve for the distortion
        filter coefficients instead of direct Gaussian elimination.
        This can speed up the computation of the metrics in case the filters
        are long. Using a value of 10 here has been shown to provide
        good accuracy in most cases and is sufficient when using this
        loss to train neural separation networks.
    zero_mean: bool, optional
        When set to True, the mean of all signals is subtracted prior
        to computation of the metrics (default: ``False``)
    clamp_db: bool, optional
        If provided, the resulting metrics are clamped to be in the range ``[-clamp_db, clamp_db]``
    load_diag: float, optional
        If provided, this small value is added to the diagonal coefficients of
        the system metrics when solving for the filter coefficients.
        This can help stabilize the metric in the case where some of the reference
        signals may sometimes be zero
    pairwise: bool, optional
        If set to True, the metrics are computed for every est/ref signals pair (default: False).
        When set to False, it is expected that ``n_channels_ref == n_channels_est``.

    Returns
    -------
    neg_cisdr: torch.Tensor
        The negative SDR of the input signal. The returned tensor has shape
        (..., n_channels_est) if ``pairwise == False`` and
        (..., n_channels_ref, n_channels_est) if ``pairwise == True``.
    """

    if filter_length == 1:
        coh = square_cosine_metrics_length_one_filter(
            ref,
            est,
            zero_mean=zero_mean,
            load_diag=None,
            pairwise=pairwise,
            with_coh_sar=False,
        )

    else:

        if zero_mean:
            est = _remove_mean(est)
            ref = _remove_mean(ref)

        # normalize along time-axis
        est = _normalize(est, dim=-1)
        ref = _normalize(ref, dim=-1)

        # compute auto-correlation and cross-correlation
        acf, xcorr = compute_stats(ref, est, length=filter_length, pairwise=pairwise)

        if load_diag is not None:
            # the diagonal factor of the Toeplitz matrix is the first
            # coefficient of the acf
            acf[..., 0] += load_diag

        # solve for the optimal filter
        if use_cg_iter is None:
            # regular matrix solver
            R_mat = toeplitz(acf)
            sol = torch.linalg.solve(R_mat, xcorr)
        else:
            # use preconditioned conjugate gradient
            sol = toeplitz_conjugate_gradient(acf, xcorr, n_iter=use_cg_iter)

        # compute the coherence
        if pairwise:
            coh = torch.einsum("...lc,...lc->...c", xcorr, sol)
        else:
            coh = torch.einsum("...l,...l->...", xcorr, sol)

    # transform to decibels
    neg_sdr = _coherence_to_neg_sdr(coh, clamp_db=clamp_db)

    return neg_sdr


def sdr_pit_loss(
    est: torch.Tensor,
    ref: torch.Tensor,
    filter_length: Optional[int] = 512,
    use_cg_iter: Optional[int] = None,
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
    load_diag: Optional[float] = None,
) -> torch.Tensor:
    return sdr(
        ref,
        est,
        filter_length=filter_length,
        use_cg_iter=use_cg_iter,
        zero_mean=zero_mean,
        clamp_db=clamp_db,
        load_diag=load_diag,
        change_sign=True,
        return_perm=False,
    )


def pairwise_sdr_loss(
    est: torch.Tensor,
    ref: torch.Tensor,
    filter_length: Optional[int] = 512,
    use_cg_iter: Optional[int] = None,
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
    load_diag: Optional[float] = None,
) -> torch.Tensor:
    """
    Compute the negative signal-to-distortion ratio (SDR).

    This function computes the SDR for all pairs of est/ref signals.

    The order of est/ref follows the convention of pytorch loss functions
    (i.e., est first).

    Parameters
    ----------
    est: torch.Tensor (..., n_channels_est, n_samples)
        The estimated signals
    ref: torch.Tensor (..., n_channels_ref, n_samples)
        The groundtruth reference signals
    filter_length: int, optional
        The length of the distortion filter allowed (default: ``512``)
    use_cg_iter: int, optional
        If provided, an iterative method is used to solve for the distortion
        filter coefficients instead of direct Gaussian elimination.
        This can speed up the computation of the metrics in case the filters
        are long. Using a value of 10 here has been shown to provide
        good accuracy in most cases and is sufficient when using this
        loss to train neural separation networks.
    zero_mean: bool, optional
        When set to True, the mean of all signals is subtracted prior
        to computation of the metrics (default: ``False``)
    clamp_db: bool, optional
        If provided, the resulting metrics are clamped to be in the range ``[-clamp_db, clamp_db]``
    load_diag: float, optional
        If provided, this small value is added to the diagonal coefficients of
        the system metrics when solving for the filter coefficients.
        This can help stabilize the metric in the case where some of the reference
        signals may sometimes be zero

    Returns
    -------
    neg_cisdr: torch.Tensor, (..., n_channels_ref, n_channels_est)
        The negative SDR of the input signal
    """
    return sdr_loss(
        est,
        ref,
        filter_length=filter_length,
        use_cg_iter=use_cg_iter,
        zero_mean=zero_mean,
        clamp_db=clamp_db,
        load_diag=load_diag,
        pairwise=True,
    )


def sdr(
    ref: torch.Tensor,
    est: torch.Tensor,
    filter_length: Optional[int] = 512,
    use_cg_iter: Optional[int] = None,
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
    load_diag: Optional[float] = None,
    return_perm: Optional[bool] = False,
    change_sign: Optional[bool] = False,
) -> Tuple[torch.Tensor, ...]:
    """
    Compute the signal-to-distortion ratio (SDR) only.

    This function computes the SDR for all pairs of est/ref signals and finds the
    permutation maximizing the SDR.

    The order of ref/est follows the convention of bss_eval (i.e., ref first).

    Parameters
    ----------
    ref: torch.Tensor (..., n_channels_est, n_samples)
        The estimated signals
    est: torch.Tensor (..., n_channels_ref, n_samples)
        The groundtruth reference signals
    filter_length: int, optional
        The length of the distortion filter allowed (default: ``512``)
    use_cg_iter: int, optional
        If provided, an iterative method is used to solve for the distortion
        filter coefficients instead of direct Gaussian elimination.
        This can speed up the computation of the metrics in case the filters
        are long. Using a value of 10 here has been shown to provide
        good accuracy in most cases and is sufficient when using this
        loss to train neural separation networks.
    zero_mean: bool, optional
        When set to True, the mean of all signals is subtracted prior
        to computation of the metrics (default: ``False``)
    clamp_db: bool, optional
        If provided, the resulting metrics are clamped to be in the range ``[-clamp_db, clamp_db]``
    load_diag: float, optional
        If provided, this small value is added to the diagonal coefficients of
        the system metrics when solving for the filter coefficients.
        This can help stabilize the metric in the case where some of the reference
        signals may sometimes be zero
    return_perm: bool, optional
        If set to True, the optimal permutation of the estimated signals is
        also returned (default: ``False``)
    change_sign: bool, optional
        If set to True, the sign is flipped and the negative SDR is returned
        (default: ``False``)

    Returns
    -------
    cisdr: torch.Tensor, (..., n_channels_est)
        The SDR of the input signal
    perm: torch.Tensor, (..., n_channels_est), optional
        The index of the corresponding reference signal.
        Only returned if ``return_perm == True``
    """

    neg_sdr = sdr_loss(
        est,
        ref,
        filter_length=filter_length,
        use_cg_iter=use_cg_iter,
        pairwise=True,
        zero_mean=zero_mean,
        clamp_db=clamp_db,
        load_diag=load_diag,
    )

    neg_sdr, perm = _solve_permutation(neg_sdr, return_perm=True)

    sdr = neg_sdr if change_sign else -neg_sdr

    if return_perm:
        return sdr, perm
    else:
        return sdr


def square_cosine_metrics(
    ref: torch.Tensor,
    est: torch.Tensor,
    filter_length: Optional[int] = 512,
    use_cg_iter: Optional[int] = None,
    zero_mean: Optional[bool] = False,
    pairwise: Optional[bool] = True,
    load_diag: Optional[float] = None,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    This function computes the square cosines of the angles between

    1. the estimate and the subspaces of shifts of each of the reference signals
    2. the estimate and the subspace of shifts of all reference signals

    The order of ref/est follows the convention of bss_eval (i.e., ref first).

    Parameters
    ----------
    ref: torch.Tensor (..., n_channels_est, n_samples)
        The estimated signals
    est: torch.Tensor (..., n_channels_ref, n_samples)
        The groundtruth reference signals
    filter_length: int, optional
        The length of the distortion filter allowed (default: ``512``)
    use_cg_iter: int, optional
        If provided, an iterative method is used to solve for the distortion
        filter coefficients instead of direct Gaussian elimination.
        This can speed up the computation of the metrics in case the filters
        are long. Using a value of 10 here has been shown to provide
        good accuracy in most cases and is sufficient when using this
        loss to train neural separation networks.
    zero_mean: bool, optional
        When set to True, the mean of all signals is subtracted prior
        to computation of the metrics (default: ``False``)
    pairwise: bool, optional
        When this flag is true, statistics are computed for all
        pairs of source/estimate. This only affects the cross-correlation.
    load_diag: float, optional
        If provided, this small value is added to the diagonal coefficients of
        the system metrics when solving for the filter coefficients.
        This can help stabilize the metric in the case where some of the reference
        signals may sometimes be zero

    Returns
    -------
    neg_cisdr: torch.Tensor, (..., n_channels_ref, n_channels_est)
        The negative SDR of the input signal
    """

    # In this case, we can compute things more efficiently
    if filter_length == 1:
        return square_cosine_metrics_length_one_filter(
            ref,
            est,
            zero_mean=zero_mean,
            pairwise=pairwise,
            load_diag=load_diag,
            with_coh_sar=True,
        )

    if zero_mean:
        est = _remove_mean(est)
        ref = _remove_mean(ref)

    # normalize along time-axis
    est = _normalize(est, dim=-1)
    ref = _normalize(ref, dim=-1)

    # compute auto-correlation and cross-correlation
    # acf.shape = (..., 2 * filter_length, n_ref, n_est)
    # xcorr.shape = (..., filter_length, n_ref, n_est)
    acf, xcorr = compute_stats_2(ref, est, length=filter_length)

    diag_indices = list(range(ref.shape[-2]))

    if load_diag is not None:
        # the diagonal factor of the Toeplitz matrix is the first
        # coefficient of the acf
        acf[..., 0, diag_indices, diag_indices] += load_diag

    # solve for the SDR
    acf_sdr = acf[..., diag_indices, diag_indices]
    acf_sdr = acf_sdr.transpose(-2, -1)
    acf_sdr = acf_sdr[..., :filter_length]

    if pairwise:
        xcorr_sdr = xcorr.transpose(-3, -2)
    else:
        xcorr_sdr = xcorr[..., :, diag_indices, diag_indices].transpose(-2, -1)

    # solve for the optimal filter
    if use_cg_iter is None:
        # regular matrix solver
        R_mat = toeplitz(acf_sdr)
        sol = torch.linalg.solve(R_mat, xcorr_sdr)
    else:
        # use preconditioned conjugate gradient
        sol = toeplitz_conjugate_gradient(acf_sdr, xcorr_sdr, n_iter=use_cg_iter)

    # pairwise coherence
    if pairwise:
        coh_sdr = torch.einsum(
            "...lc,...lc->...c", xcorr_sdr, sol
        )  # (..., n_ref, n_est)
    else:
        coh_sdr = torch.einsum("...l,...l->...", xcorr_sdr, sol)

    # solve the coefficients for the SIR
    xcorr = xcorr.reshape(xcorr.shape[:-3] + (-1,) + xcorr.shape[-1:])
    if use_cg_iter is None:
        R_mat = block_toeplitz(acf)
        sol = torch.linalg.solve(R_mat, xcorr)
    else:
        if pairwise:
            x0 = sol
            x0 = sol.transpose(-3, -2)
            x0 = x0.reshape(x0.shape[:-3] + (-1,) + x0.shape[-1:])
        else:
            x0 = None
        sol = block_toeplitz_conjugate_gradient(acf, xcorr, n_iter=use_cg_iter, x=x0)
    coh_sar = torch.einsum("...lc,...lc->...c", xcorr, sol)

    if pairwise:
        coh_sdr, coh_sar = torch.broadcast_tensors(coh_sdr, coh_sar[..., None, :])

    return coh_sdr, coh_sar


def _base_metrics_bss_eval(
    coh_sdr: torch.Tensor, coh_sar: torch.Tensor, clamp_db: Optional[float] = None
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Helper function that converts the square cosine metrics to SDR/SIR/SAR
    """

    coh_sir = coh_sdr / coh_sar

    neg_sdr = _coherence_to_neg_sdr(coh_sdr, clamp_db=clamp_db)
    neg_sir = _coherence_to_neg_sdr(coh_sir, clamp_db=clamp_db)
    neg_sar = _coherence_to_neg_sdr(coh_sar, clamp_db=clamp_db)

    return neg_sdr, neg_sir, neg_sar


def bss_eval_sources(
    ref: torch.Tensor,
    est: torch.Tensor,
    filter_length: Optional[int] = 512,
    use_cg_iter: Optional[int] = None,
    zero_mean: Optional[bool] = False,
    clamp_db: Optional[float] = None,
    compute_permutation: Optional[bool] = True,
    load_diag: Optional[float] = None,
) -> torch.Tensor:
    """
    This function computes the SDR, SIR, and SAR for the input reference and estimated
    signals.

    The order of ref/est follows the convention of bss_eval (i.e., ref first).

    Parameters
    ----------
    ref: torch.Tensor (..., n_channels_est, n_samples)
        The estimated signals
    est: torch.Tensor (..., n_channels_ref, n_samples)
        The groundtruth reference signals
    filter_length: int, optional
        The length of the distortion filter allowed (default: ``512``)
    use_cg_iter: int, optional
        If provided, an iterative method is used to solve for the distortion
        filter coefficients instead of direct Gaussian elimination.
        This can speed up the computation of the metrics in case the filters
        are long. Using a value of 10 here has been shown to provide
        good accuracy in most cases and is sufficient when using this
        loss to train neural separation networks.
    zero_mean: bool, optional
        When set to True, the mean of all signals is subtracted prior
        to computation of the metrics (default: ``False``)
    clamp_db: bool, optional
        If provided, the resulting metrics are clamped to be in the range ``[-clamp_db, clamp_db]``
    compute_permutation: bool, optional
        When this flag is true, the optimal permutation of the estimated signals
        to maximize the SIR is computed (default: ``True``)
    load_diag: float, optional
        If provided, this small value is added to the diagonal coefficients of
        the system metrics when solving for the filter coefficients.
        This can help stabilize the metric in the case where some of the reference
        signals may sometimes be zero

    Returns
    -------
    sdr: torch.Tensor, (..., n_channels_est)
        The signal-to-distortion-ratio (SDR)
    sir: torch.Tensor, (..., n_channels_est)
        The signal-to-interference-ratio (SIR)
    sar: torch.Tensor, (..., n_channels_est)
        The signal-to-interference-ratio (SIR)
    perm: torch.Tensor, (..., n_channels_est)
        The index of the corresponding reference signal (only when
        ``compute_permutation == True``)
    """

    if not compute_permutation:
        assert ref.shape[-2] == est.shape[-2], (
            "For non-pairwise computation, the number of reference"
            " and estimated signals must be the same"
        )

    coh_sdr, coh_sar = square_cosine_metrics(
        ref,
        est,
        filter_length=filter_length,
        use_cg_iter=use_cg_iter,
        zero_mean=zero_mean,
        pairwise=compute_permutation,
        load_diag=load_diag,
    )

    neg_sdr, neg_sir, neg_sar = _base_metrics_bss_eval(
        coh_sdr, coh_sar, clamp_db=clamp_db
    )

    if compute_permutation:
        neg_sir, neg_sdr, neg_sar, perm = _solve_permutation(
            neg_sir, neg_sdr, neg_sar, return_perm=True
        )
        return -neg_sdr, -neg_sir, -neg_sar, perm
    else:
        return -neg_sdr, -neg_sir, -neg_sar



import pathlib

from setuptools import find_packages, setup

here = pathlib.Path(__file__).parent.resolve()

try:
    long_description = (here / "README.md").read_text(encoding="utf-8")
except Exception:
    long_description = ""

setup(
    name="fast_bss_eval",
    version="0.1.1",
    description="Package for fast computation of BSS Eval metrics for source separation",
    long_description=long_description,
    long_description_content_type="text/markdown",  # text/plain, text/x-rst, text/markdown
    url="https://github.com/fakufaku/fast_bss_eval",
    author="Robin Scheibler",
    author_email="fakufaku@gmail.com",
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        # Pick your license as you wish
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate you support Python 3. These classifiers are *not*
        # checked by 'pip install'. See instead 'python_requires' below.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
    ],
    # keywords='sample, setuptools, development',
    packages=find_packages(exclude=["contrib", "docs", "tests*"]),
    python_requires=">=3.6, <4",
    install_requires=[
        "numpy",
        "scipy",
    ],
    extras_require={
        "all": [],
    },
    # package_data={
    #     'sample': ['package_data.dat'],
    # },
    # data_files=[('my_data', ['data/data_file'])],
    # entry_points={
    #     'console_scripts': [
    #         'sample=sample:main',
    #     ],
    # },
    project_urls={  # Optional
        "Bug Reports": "https://github.com/fakufaku/fast_bss_eval/issues",
        "Source": "https://github.com/fakufaku/fast_bss_eval/",
    },
)
